import os
import sqlite3
import re
import codecs

word_boundary = re.compile("\W+", re.U)

conn = sqlite3.connect("questions.db") # Database of questions
c = conn.cursor()
c.execute('CREATE TABLE IF NOT EXISTS words (id INTEGER PRIMARY KEY, word varchar(255) NOT NULL UNIQUE, frequency UNSIGNED NOT NULL)')
c.execute('CREATE TABLE IF NOT EXISTS word_to_question (word_id UNSIGNED NOT NULL, question_id INT UNSIGNED NOT NULL, PRIMARY KEY(word_id, question_id))')
c.execute('CREATE TABLE IF NOT EXISTS questions (id INTEGER PRIMARY KEY, start_point INT UNSIGNED NOT NULL, length INT UNSIGNED NOT NULL, question TEXT)')
c.execute('CREATE TABLE IF NOT EXISTS config (key VARCHAR(255) NOT NULL PRIMARY KEY, value VARCHAR(255))')
c.execute('INSERT OR IGNORE INTO config VALUES("version", "00001")')
c.execute('INSERT OR IGNORE INTO config VALUES("initialised", "")')
conn.commit()

def add_pages(pages):
    """Add the specified pages to the database"""

    # The "super log" is a superset of all the other log files
    super_log = codecs.open("super-log.txt", 'wb+', encoding="utf-8")

    # Matches a message
    MESSAGE = re.compile("\[([0-9]{2}):([0-9]{2})\] <(.*?)> (.+ .+ .+)")
    # "person: message" or "person, message"
    MESSAGE_TO = re.compile("^(\S+)[:,]\s", re.U)
    # "person: message" or "person, message", "!message | person", "!message > person"
    MESSAGE_IGNORE = re.compile("\S*\s*[:|,>]\s+.*", re.U)
    # "why does X" "Y?", "do you know when Z", etc.
    QUESTION_HINTS = re.compile("\?($|\s)|(^|\s)why($|\s)|(^|\s)what($|\s)|(^|\s)where($|\s)", re.U)

    # What encoding do log files use?  Evidence suggests a nasty mess of UTF-8 and Latin-1
    SOURCE_ENCODING="latin1"

    RECENT = 12 * 60 # "recent" means "within 12 hours"

    today = 0 # time when the current log file was started
    ignore_until = {} # everyone that's ever said anything, and when 
    latest_questions = {} # Question text and start points
    answer_endpoints = {} # Question end-points
    active_questions = [ "", "", "", "", "", "", "", "", "", "" ] # Print 10 lines after each question
    questions_start = 0 # value of superlog.len() for the first question to save
    questions_end = 0 # value of superlog.len() for the last question to save

    def add_question(question_data, endpoint):
        """Add a single question to the database"""
        # Ignore questions that started on the first day
        if (questions_start > endpoint or not questions_start): return
        (question, startpoint) = question_data

        # Ignore questions that started on the last day
        if (questions_end and questions_end <= startpoint): return

        c.execute("INSERT INTO questions(start_point, length, question) VALUES (?,?,?)", (startpoint, endpoint-startpoint, question))

        question_id = c.execute("SELECT id FROM questions WHERE start_point=?", (startpoint,)).fetchone()[0]

        words = word_boundary.split(question)

        previous_words = {}
        for word in words:
            if word not in previous_words:
                c.execute("INSERT OR IGNORE INTO words(word, frequency) VALUES(?,0)", (word,))
                c.execute("UPDATE words SET frequency=frequency+1 WHERE word=?", (word,))
                c.execute("INSERT INTO word_to_question(word_id, question_id) SELECT id, ? FROM words WHERE word=?", (question_id, word,))
                previous_words[word] = None

    for n in range(0, len(pages)):

        # Do not insert the last page, or reinsert an old page
        if len(pages)-1 == n or 1 == c.execute("SELECT COUNT(*) FROM config WHERE key=?", ('ADDED ' + pages[n],)).fetchone():
            questions_end = super_log.tell()
        else:
            questions_end = 0
            c.execute("INSERT INTO config VALUES(?, 'yes')", ('ADDED ' + pages[n],))

        print pages[n], "...",

        for line in codecs.open(pages[n], encoding=SOURCE_ENCODING):
            match = MESSAGE.match(line)


            if match is not None and MESSAGE_IGNORE.match(match.group(4)) is None:

                time = today + int(match.group(1)) * 60 + int(match.group(2))
                name = match.group(3)
                if name in ignore_until: # This person has said something before

                    # If this person hasn't spoken recently, add their previous question
                    if ignore_until[name] < time:
                        if name in answer_endpoints:
                            add_question(latest_questions[name], answer_endpoints[name])
                            del answer_endpoints[name]
                        else: pass
                        latest_questions[name] = [ match.group(4), super_log.tell() ]

                    # If this is a follow-up question, add it to the question text
                    elif QUESTION_HINTS.search(match.group(4)) is not None:
                        latest_questions[name][0] = latest_questions[name][0] + "\n" + match.group(4)
                    else:
                        answer_endpoints[name] = super_log.tell() + len(line)

                else: # This person hasn't said anything before
                    latest_questions[name] = [ match.group(4), super_log.tell() ]

                ignore_until[name] = time + RECENT

                # For "person: message", note that this person's question is still live
                match = MESSAGE_TO.search(line)
                if match is not None and match.group(1) in ignore_until:
                    ignore_until[match.group(1)] = time + RECENT
                    answer_endpoints[name] = super_log.tell() + len(line)
                else: pass

                # Questions remain live for at least ten lines after the message is sent
                active_questions.insert(0, name)
                answer_endpoints[active_questions.pop()] = super_log.tell() + len(line)

            else: pass

            super_log.write(line)

        today += 24 * 60

        conn.commit()

        print "done."

        questions_start = super_log.tell()

    # Add the last question that everybody asked
    for name in latest_questions:
        if name in answer_endpoints:
            add_question(latest_questions[name], answer_endpoints[name])

    conn.commit()

def get_questions(text):
    """Get a list of questions from the database"""

    words = word_boundary.split(text)

    c.execute("""
    SELECT questions.start_point, questions.question
      FROM words, word_to_question, questions
     WHERE words.word in (?""" + (",?" * ( len(words) - 1)) + """)
       AND word_to_question.    word_id =     words.id
       AND word_to_question.question_id = questions.id
     GROUP BY questions.start_point
     ORDER BY sum(1.0/frequency) desc, questions.start_point desc
     """
              , words)

    return c

def get_questions_subwords(text):
    """Get a list of questions from the database, allowing sub-words (e.g. 'nic' matches 'frobnicate')"""

    words = word_boundary.split(text)

    c.execute("""
    SELECT questions.start_point, questions.question
      FROM words, word_to_question, questions
     WHERE (""" +     ('words.word LIKE ?\n        OR ' * len(words)) + """1=0)
       AND word_to_question.    word_id =     words.id
       AND word_to_question.question_id = questions.id
     GROUP BY questions.start_point
     ORDER BY sum(1.0/frequency) desc, questions.start_point desc
     """
              , map(lambda x: '%' + x + '%', words))

    return c

def get_question(start_point):
    """Get the text for a specific question from the database"""

    c.execute("SELECT length FROM questions WHERE start_point = ?", (start_point,))
    row = c.fetchone()

    if row is not None:
        super_log = codecs.open("super-log.txt", 'rb', encoding="utf-8")
        super_log.seek(start_point)
        return super_log.read(row[0]);
    else:
        return None

def initialise(start_date, end_date):

    from calendar import monthrange

    (start_year, start_month, start_day) = start_date
    (  end_year,   end_month,   end_day) =   end_date

    command="curl --compressed -C - --create-dirs"
    run_command=False
    url_base=" http://{irclogs.ubuntu.com/%d/%02d/%02d/%%23ubuntu.txt} -o irclogs.ubuntu.com/%d/%02d/%02d/#ubuntu.txt"
    file_base="irclogs.ubuntu.com/%d/%02d/%02d/#ubuntu.txt"
    pages=[]

    for year in range(start_year, end_year+1):
        if year == end_year:
            month_range   = range(start_month, end_month+1)
        else: month_range = range(start_month, 12)

        for month in month_range:
            if year == end_year and month == end_month:
                day_range = range(start_day, end_day)
            else:
                day_range = range(start_day, monthrange(year, month+1)[1]+1)

            for day in day_range:
                if not os.path.exists(file_base % ( year, month+1, day )):
                    command += url_base % ( year, month+1, day, year, month+1, day )
                    run_command=True
                pages.append(file_base % ( year, month+1, day ))

            start_day = 1
        start_month = 0

    if run_command:
        os.system(command)

    add_pages(pages)

    c.execute('UPDATE config SET value="yes" WHERE key="initialised"')
    conn.commit()

def is_initialised():
    c.execute('SELECT value FROM config WHERE key="initialised"')
    return "yes" == c.fetchone()[0]
